#! /usr/bin/env python

# Panflute
# Copyright (C) 2009 Paul Kuliniewicz <paul@kuliniewicz.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.

"""
GObject-based interface to the Panflute MPRIS daemon.

This module encapsulates the details of talking to the Panflute MPRIS
daemon, exposing the various interfaces through a single object.
"""

import panflute.mpris

import dbus
import gobject
import gtk


class Player (gobject.GObject):
    """
    Proxy object for the Panflute MPRIS daemon.

    This caches the current state exposed by the daemon, eliminating some of
    the communications that would otherwise be needed.  It particular, it makes
    it possible to create widgets that can immediately access the current
    state of the player, without needing one or more round trips to get that
    information from the daemon.
    """


    ##########################################################################
    

    __gproperties__ = {
        "can-pause": (gobject.TYPE_BOOLEAN,
                      "can-pause",
                      "Whether the player can currently pause",
                      False,
                      gobject.PARAM_READABLE),

        "can-go-next": (gobject.TYPE_BOOLEAN,
                        "can-go-next",
                        "Whether the player can currently go to the next song",
                        False,
                        gobject.PARAM_READABLE),

        "can-go-previous": (gobject.TYPE_BOOLEAN,
                            "can-go-previous",
                            "Whether the player can currently go to the previous song",
                            False,
                            gobject.PARAM_READABLE),

        "state": (gobject.TYPE_UINT,
                  "state",
                  "Current playback state",
                  panflute.mpris.STATE_MIN, panflute.mpris.STATE_MAX,
                  panflute.mpris.STATE_STOPPED,
                  gobject.PARAM_READABLE),

        "title": (gobject.TYPE_STRING,
                  "title",
                  "Title of the current song",
                  None,
                  gobject.PARAM_READABLE),

        "artist": (gobject.TYPE_STRING,
                   "artist",
                   "Artist performing the current song",
                   None,
                   gobject.PARAM_READABLE),

        "album": (gobject.TYPE_STRING,
                  "album",
                  "Album the current song is from",
                  None,
                  gobject.PARAM_READABLE),

        "duration": (gobject.TYPE_UINT,
                     "duration",
                     "Length of the current song, in milliseconds",
                     0, gobject.G_MAXUINT,
                     0,
                     gobject.PARAM_READABLE),

        "elapsed": (gobject.TYPE_UINT,
                    "elapsed",
                    "Position within the current song, in milliseconds",
                    0, gobject.G_MAXUINT,
                    0,
                    gobject.PARAM_READABLE),

        "rating": (gobject.TYPE_UINT,
                   "rating",
                   "Rating of the current song",
                   0, 5,
                   0,
                   gobject.PARAM_READABLE),

        "art": (gobject.TYPE_OBJECT,
                "art",
                "Artwork associated with the current song",
                gobject.PARAM_READABLE),

        "art-file": (gobject.TYPE_STRING,
                     "art-file",
                     "File name of the artwork associated with the current song",
                     None,
                     gobject.PARAM_READABLE)
    }


    from panflute.util import log


    def __init__ (self):
        gobject.GObject.__init__ (self)

        self.__props = {
            "can-pause":       False,
            "can-go-next":     False,
            "can-go-previous": False,
            "state":           panflute.mpris.STATE_STOPPED,
            "title":           None,
            "artist":          None,
            "album":           None,
            "duration":        0,
            "elapsed":         0,
            "rating":          0,
            "art":             None,
            "art-file":        None
        }

        bus = dbus.SessionBus ()
        proxy = bus.get_object ("org.mpris.panflute", "/Player")
        self.__player = dbus.Interface (proxy, panflute.mpris.INTERFACE)
        self.__player_ex = dbus.Interface (proxy, "org.kuliniewicz.Panflute")

        self.__dbus_handlers = [
            self.__player.connect_to_signal ("CapsChange", self.__caps_change_cb),
            self.__player.connect_to_signal ("StatusChange", self.__status_change_cb),
            self.__player.connect_to_signal ("TrackChange", self.__track_change_cb),
            self.__player_ex.connect_to_signal ("PositionChange", self.__position_change_cb)
        ]

        self.__player.GetCaps (reply_handler = self.__caps_change_cb,
                               error_handler = self.log.warn)
        self.__player.GetStatus (reply_handler = self.__status_change_cb,
                                 error_handler = self.log.warn)
        self.__player.GetMetadata (reply_handler = self.__track_change_cb,
                                   error_handler = self.log.warn)
        self.__player.PositionGet (reply_handler = self.__position_change_cb,
                                   error_handler = self.log.warn)


    def __del__ (self):
        self.log.debug ("Inside destructor")

        for handler in self.__dbus_handlers:
            handler.disconnect ()
        self.__dbus_handlers = []


    def do_get_property (self, property):
        return self.__props[property.name]


    def __set_property (self, name, value):
        """
        Set the value of a property, issuing the notify signal if the value
        does in fact change.

        This is distinct from the set_property provided by gobject in that
        it prevents outsiders from trying to set properties.
        """

        if self.__props[name] != value:
            self.__props[name] = value
            self.notify (name)


    def __caps_change_cb (self, caps):
        """
        Update the capabilities properties with the current caps.
        """

        self.__set_property ("can-pause", (caps & panflute.mpris.CAN_PAUSE) != 0)
        self.__set_property ("can-go-next", (caps & panflute.mpris.CAN_GO_NEXT) != 0)
        self.__set_property ("can-go-previous", (caps & panflute.mpris.CAN_GO_PREV) != 0)


    def __status_change_cb (self, status):
        """
        Update the status properties with the current status.
        """

        self.__set_property ("state", status[panflute.mpris.STATUS_STATE])


    def __track_change_cb (self, metadata):
        """
        Update the properties with the latest metadata.
        """

        old_title = self.props.title
        old_artist = self.props.artist
        old_album = self.props.album

        self.__set_property ("title", metadata.get ("title", None))
        self.__set_property ("artist", metadata.get ("artist", None) or None)
        self.__set_property ("album", metadata.get ("album", None) or None)
        self.__set_property ("duration", metadata.get ("mtime", 0))
        self.__set_property ("rating", metadata.get ("rating", 0))

        art = None
        art_file = None
        if metadata.has_key ("arturl"):
            if metadata["arturl"].startswith ("file://"):
                try:
                    art_file = metadata["arturl"][7:]
                    art = gtk.gdk.pixbuf_new_from_file (art_file)
                except gobject.GError:
                    self.log.info ("Failed to load {0}".format (metadata["arturl"]))
            else:
                self.log.info ("Don't know how to read {0}".format (metadata["arturl"]))

        # Yes, set art-file even if loading art failed -- maybe some other
        # program (e.g. pynotify) will be able to open it.
        self.__set_property ("art", art)
        self.__set_property ("art-file", art_file)

        if self.props.title != old_title or self.props.artist != old_artist or self.props.album != old_album:
            self.emit ("song-changed")


    def __position_change_cb (self, position):
        """
        Update the elapsed property with the latest value.
        """

        self.__set_property ("elapsed", position)


    def pause (self):
        """
        Play or pause playback.

        This pedantically figures out which MPRIS method should be called to
        actually play or pause (or stop) playback, since Pause won't necessarily
        actually toggle playback for some MPRIS players (e.g. Audacious won't
        start playing in response to a Pause if playback is stopped).
        """

        if self.props.state == panflute.mpris.STATE_PLAYING:
            if self.props.can_pause:
                method = self.__player.Pause
            else:
                method = self.__player.Stop
        elif self.props.state == panflute.mpris.STATE_PAUSED:
            method = self.__player.Pause
        else:
            method = self.__player.Play

        method (reply_handler = lambda: None,
                error_handler = self.log.warn)


    def next (self):
        """
        Advance to the next song.
        """

        self.__player.Next (reply_handler = lambda: None,
                            error_handler = self.log.warn)


    def previous (self):
        """
        Go back to the previous song.
        """

        self.__player.Prev (reply_handler = lambda: None,
                            error_handler = self.log.warn)


    def rate_song (self, rating):
        """
        Rate the current song.
        """

        self.__player_ex.SetMetadata ("rating", rating,
                                      reply_handler = lambda: None,
                                      error_handler = self.log.warn)


gobject.type_register (Player)
gobject.signal_new ("song-changed", Player,
                    gobject.SIGNAL_RUN_LAST,
                    gobject.TYPE_NONE,
                    ())
